import copy, loady, functools, sys
from . import aliases, importer
from .. project import data_maker
from .. util import log

from .. layout.geometry import make_strip_coord_map_multi, make_matrix_coord_map_multi

RESERVED_PROPERTIES = 'name', 'data'

ISNT_GIT_PATH_ERROR = """\
Because the --isolate flag is set, all paths must start with //git.
Your path was %s."""

MULTI_HANDLERS = {
    'bibliopixel.layout.strip.Strip': make_strip_coord_map_multi,
    'bibliopixel.layout.matrix.Matrix': make_matrix_coord_map_multi
}


def make_animation(layout, animation, run=None):
    animation = aliases.resolve(animation)
    reserved = {p: animation.pop(p, None) for p in RESERVED_PROPERTIES}
    animation_obj = importer.make_object(layout, **animation)

    # Add the reserved properties back in.
    for k, v in reserved.items():
        animation[k] = v
        (v is not None) and setattr(animation_obj, k, v)

    animation_obj.set_runner(run)
    return animation_obj


def extend_path(path):
    if not path:
        return

    if aliases.ISOLATE:
        if not all(x.startswith('//git/') for x in path.split(':')):
            raise ValueError(ISNT_GIT_PATH_ERROR % path)

    loady.sys_path.extend(path)


def raise_if_unknown(items, name, value):
    if items:
        msg = ', '.join('"%s"' % s for s in sorted(items))
        s = '' if len(items) == 1 else 's'
        raise ValueError('Unknown %s%s for %s: %s' % (name, s, value, msg))


def raise_if_unknown_attributes(items, name, value):
    value = '%s %s' % (name, value.__class__.__name__)
    raise_if_unknown(items, 'attribute', value)


def project_to_animation(desc, default=None):
    project = copy.deepcopy(desc)
    default = default or {}

    def resolve_aliases(name):
        pr = aliases.resolve(project.pop(name, {}))
        de = aliases.resolve(default.get(name, {}))
        return dict(de, **pr)

    animation = resolve_aliases('animation')
    driver = resolve_aliases('driver')
    layout = resolve_aliases('layout')

    drivers = project.pop('drivers', [])
    maker = project.pop('maker', {})
    path = project.pop('path', default.get('path'))
    run = project.pop('run', {})

    raise_if_unknown(project, 'section', 'project')

    if not animation:
        raise ValueError('There was no "animation" section in the project')

    if not layout:
        raise ValueError('There was no "layout" section in the project')

    if not (driver or drivers):
        raise ValueError(
            'The project has neither a "driver" nor a "drivers" section')

    extend_path(path)
    maker = data_maker.Maker(**(maker or {}))
    make_object = functools.partial(importer.make_object, maker=maker)

    driver_objects = make_drivers(driver, drivers, make_object)

    gen_coord_map = layout.pop('gen_coord_map', None)

    if 'coord_map' not in layout and gen_coord_map:
        typename = layout['typename']
        if typename not in MULTI_HANDLERS:
            raise ValueError('There is currently no available multi-map builder for {}'.format(typename))
        gen_multi = MULTI_HANDLERS[typename]
        if isinstance(gen_coord_map, dict):
            layout['coord_map'] = gen_multi(**gen_coord_map)
        elif isinstance(gen_coord_map, list):
            layout['coord_map'] = gen_multi(gen_coord_map)

    layout_object = make_object(driver_objects, **layout)

    return make_animation(layout_object, animation, run)


def make_drivers(driver, drivers, make_object):
    if not drivers:
        return [make_object(**aliases.resolve(driver))]

    if driver:
        # driver is a default for each driver.
        drivers = [dict(driver, **d) for d in drivers]

    return [make_object(**aliases.resolve(d)) for d in drivers]


def read_project(location, threaded=True, default=None):
    project = loady.data.load(location, use_json=True)
    if threaded is not None:
        project.setdefault('run', {})['threaded'] = threaded
    return project_to_animation(project, default)



import re
from .. util import log
from . importer import import_symbol
from . import alias_lists

ALIAS_MARKER = '$'
SEPARATORS = re.compile(r'[./#]|[^./#]+')
DEPRECATE_OLD_ALIASES = False

# This can set from the commamnd line with --isolate
ISOLATE = False


def resolve_one(part):
    if not part.startswith(ALIAS_MARKER):
        return part

    p = alias_lists.get_alias(part[1:], ISOLATE)
    if p:
        return p

    raise ValueError('Do not understand alias %s' % part)


def resolve(value):
    if isinstance(value, str):
        typename, value = value, {}
    else:
        typename = value.get('typename')
        if not typename:
            return value

    alias = alias_lists.get_alias(typename, ISOLATE)

    if alias:
        if DEPRECATE_OLD_ALIASES:
            log.warning('Aliases that do not start with $ are deprecated')

    else:
        parts = SEPARATORS.findall(typename)
        alias = ''.join(resolve_one(p) for p in parts)

    value['typename'] = alias
    return value



import logging, sys
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

LOG_NAMES = {'debug': DEBUG, 'info': INFO, 'warning': WARNING, 'error': ERROR,
             'critical': CRITICAL}


class InfoFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno in (DEBUG, INFO)


def _new_custom_logger(name='BiblioPixel',
                       fmt='%(levelname)s - %(module)s - %(message)s'):
    logger = logging.getLogger(name)
    formatter = logging.Formatter(fmt=fmt)

    if not logger.handlers:
        logger.setLevel(INFO)
        h1 = logging.StreamHandler(sys.stdout)
        h1.setLevel(DEBUG)
        h1.addFilter(InfoFilter())
        h1.setFormatter(formatter)

        h2 = logging.StreamHandler(sys.stderr)
        h2.setLevel(WARNING)
        h2.setFormatter(formatter)

        logger.addHandler(h1)
        logger.addHandler(h2)

    else:  # pragma: no cover
        pass

    return logger


def set_log_level(level):
    if isinstance(level, str):
        level = LOG_NAMES[level]

    logger.setLevel(level)


logger = _new_custom_logger()

setLogLevel = set_log_level

debug, info, warning, error, critical, exception = (
    logger.debug, logger.info, logger.warning, logger.error, logger.critical,
    logger.exception)



"""
Run a project description file.
"""

import loady, json, os
from . import common_flags, simpixel
from .. util import log


def run(args):
    common_flags.extend_path(args)
    if args.json:
        desc = args.name
    else:
        desc = args.name and loady.data.load(args.name, True)

    animation = common_flags.make_animation(args, desc or {})

    if args.simpixel:
        simpixel.open_simpixel(args.simpixel)
    elif args.s:
        simpixel.open_simpixel()

    try:
        animation.start()
    except KeyboardInterrupt:
        print('\nTermination requested by user.')
        animation.cleanup()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = 'Run specified BiblioPixel project from file or URL.'

    common_flags.add_project_flags(parser)

    parser.add_argument(
        'name', nargs='?',
        help='Path project file - can be a URL or file system location',
        default='')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Enter JSON directly as a command line argument.')



from .. project import aliases, project
from . import animation
import copy, traceback
from .. util import log


class Collection(animation.BaseAnimation):
    def __init__(self, layout, animations=None, no_fail=False, **kwds):
        super().__init__(layout, **kwds)
        self.animations = [self._make_animation(i, no_fail) for i in animations or []]
        self.index = 0
        self.internal_delay = 0  # never wait

    # Override to handle all the animations
    def cleanup(self, clean_layout=True):
        self.state = animation.STATE.canceled
        for a in self.animations:
            if a:
                a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        # DEPRECATED.
        anim.set_runner(kwds)
        self.animations.append(anim)

    def pre_run(self):
        self.index = -1

    @property
    def current_animation(self):
        if (0 <= self.index < len(self.animations)):
            return self.animations[self.index]
        else:
            return None

    def _make_animation(self, a, no_fail=False):
        if isinstance(a, str):
            animation = a
            run = None

        elif isinstance(a, dict):
            animation = a.get('animation')
            if animation:
                # Looks like {'animation': ..., 'run': }
                run = a.get('run')
            else:
                # It's an animation itself.
                animation, run = a, None

        else:
            animation, run = a

        assert animation, '%s:%s:%s' % (a, animation, run)
        try:
            return project.make_animation(self.layout, animation, run)
        except:
            log.error(traceback.format_exc())
            return None



import json, os
from .. project import project

"""Common command line arguments for run and demo."""


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(dirname(dirname(abspath(
        __file__)))), 'bibliopixel', 'VERSION')
    return open(filename).read().strip()


COMPONENTS = 'driver', 'layout', 'animation'
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
LOG_LEVELS = ('debug', 'info', 'warning', 'error', 'critical')
ENABLE_PRESETS = False

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use loady-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

LOGLEVEL_HELP = """\
Set what level of events to log. Higher log levels print less."""

VERBOSE_HELP = """\
If this is set, then errors are reported with a full stack trace.
If not set, just the exception message is printed.
"""

VERSION = _get_version()
VERSION_HELP = """\
Print the current version number of BiblioPixel (%s).
""" % VERSION

PRESET_HELP = """Filenames for preset library"""
ISOLATE_HELP = """\
Run BiblioPixel in isolated mode, where it cannot see your local files.
This means that it will not see any local Python classes in your directories
and it won't see your local aliases.

Running your project in isolated mode help makes sure that your project will
work on other machines without modification.
"""


def add_common_flags(parser):
    parser.add_argument(
        '--loglevel', choices=LOG_LEVELS, default='info', help=LOGLEVEL_HELP)
    parser.add_argument(
        '--verbose', '-v', action='store_true', help=VERBOSE_HELP)
    parser.add_argument(
        '--version', action='store_true', help=VERSION_HELP)
    parser.add_argument(
        '-x', '--isolate', action='store_true', help=ISOLATE_HELP)

    if ENABLE_PRESETS:
        parser.add_argument(
            '--presets', help=PRESET_HELP, default=PRESET_LIBRARY_DEFAULT)


def add_project_flags(parser):
    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')

    parser.add_argument(
        '-d', '--driver', default='simpixel',
        help='Default driver type if no driver is specified')

    parser.add_argument(
        '-l', '--layout', default='matrix',
        help='Default layout class if no layout is specified')

    parser.add_argument(
        '-n', '--numpy', action='store_true',
        help='Use numpy if it available.')

    parser.add_argument(
        '-p', '--path', default=None, help=PATH_HELP)

    parser.add_argument(
        '-s', action='store_true', help='Run SimPixel at the default URL')

    parser.add_argument(
        '--simpixel', help='Run SimPixel at a specific URL')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')

    parser.add_argument(
        '-b', '--brightness', default=None,
        help='Override project brightness value')


def make_project_flags(args):
    def get_value(name):
        value = getattr(args, name)
        if not value:
            return {}

        if '{' in value:
            return json.loads(value)

        return {'typename': value}

    project_flags = {name: get_value(name) for name in COMPONENTS}
    if args.ledtype:
        project_flags['driver']['ledtype'] = args.ledtype

    if args.brightness:
        project_flags['layout']['brightness'] = int(args.brightness)

    if args.numpy:
        project_flags['maker'] = {'use_numpy': True}

    return project_flags


def make_animation(args, desc):
    project_flags = make_project_flags(args)
    return project.project_to_animation(desc, project_flags)


def extend_path(args):
    if args.isolate:
        path = args.path
    else:
        path = os.getcwd()
        if args.path:
            path += ':' + args.path
    project.extend_path(path)



from . import collection
import random as rand


class Sequence(collection.Collection):
    def __init__(self, layout, random=False, **kwds):
        super().__init__(layout, **kwds)
        self.random = random
        self.indices = list(range(0, len(self.animations)))
        self.sub_index = 0

    def pre_run(self):
        if self.random:
            rand.shuffle(self.indices)
        self.sub_index = 0

    def step(self, amt=1):
        self.sub_index += 1

        if self.sub_index >= len(self.animations):
            if self.runner.until_complete:
                self.completed = True
            else:
                rand.shuffle(self.indices)
                self.sub_index = 0

        self.index = self.indices[self.sub_index]

        if not self.completed and self.animations:
            self.current_animation.run_all_frames()
